/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.transfer.reliable.service;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.rft.generated.OverallStatus;
import org.globus.rft.generated.RFTFaultResourcePropertyType;
import org.globus.rft.generated.RequestStatusTypeEnumeration;
import org.globus.rft.generated.TransferType;
import org.globus.rft.generated.RequestStatusType;
import org.globus.util.I18n;

import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;

import org.globus.transfer.reliable.service.database.RFTDatabasePoller;
import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;

import org.globus.transfer.reliable.service.client.Client;
import org.globus.transfer.reliable.service.client.URLExpanderClient;
import org.globus.transfer.reliable.service.client.ClientFactory;

/**
 * Does the actual transfer
 *  
 */
public class TransferWork implements Runnable {

    private TransferType transferJob;

    private int status = -1;
    private int attempts;
    private int requestId = 0;
    private int maxAttempts = 0;

    private ReliableFileTransferDbAdapter dbAdapter =
    ReliableFileTransferDbAdapter.getDbAdapter();

    private static Log logger = LogFactory.getLog(TransferWork.class.getName());

    private static I18n i18n = I18n.getI18n("org.globus.transfer.reliable.errors",
            TransferWork.class.getClassLoader());
    
    private OverallStatus overallStatus;

    private ResourceProperty overallStatusRP;
    private ResourceProperty requestStatusTypeRP;

    private ResourcePropertySet rpSet;

    private RequestStatusType requestStatusType;

    private RFTFaultResourcePropertyType fault;
    private RFTResourceManager resourceManager = null;

    /**
     * Constructor for the TransferWork object
     * 
     * @param requestId
     * @param rpSet
     * @param transferJob
     */
    public TransferWork(int requestId, ResourcePropertySet rpSet,
                        TransferType transferJob)
    {
        this.requestId = requestId;
        this.rpSet = rpSet;
        this.transferJob = transferJob;
        this.status = -1;
        this.attempts = transferJob.getAttempts().intValue();
        this.overallStatusRP = 
            this.rpSet.get(RFTConstants.OVERALL_STATUS_RESOURCE);
        this.overallStatus = (OverallStatus) this.overallStatusRP.get(0);
        this.requestStatusTypeRP = 
            this.rpSet.get(RFTConstants.REQUEST_STATUS_RESOURCE);
        this.requestStatusType = (RequestStatusType) 
        this.requestStatusTypeRP.get(0);
    }

    /**
     * @param transferJob
     */
    public void setTransferJob(TransferType transferJob) {
        this.transferJob = transferJob;
    }

    /**
     * Gets the status attribute of the TransferWork object
     * 
     * @return The status value
     */
    public int getStatus() {
        return this.status;
    }

    public void setResourceManager(RFTResourceManager resourceManager) 
        throws RftDBException {
        this.resourceManager = resourceManager;
        this.maxAttempts = this.resourceManager.getMaxAttempts();  
    }
    
    /**
     * generates notification of status changes of a request
     * 
     * @param status
     */
    private synchronized void statusChanged(int status) {
        try {
            this.status = status;
            if (logger.isDebugEnabled()) {
                logger.debug(this.getTransferIdentifiers()
                             + "status changed called with status: " 
                             + this.status);
            }
            this.transferJob.setStatus(RFTUtil.mapStatus(this.status));
            this.resourceManager.updateTransferJob( 
                    this.transferJob.getTransferId().intValue(), 
                    this.status, 
                    this.attempts);
            //this.dbAdapter.update(this.requestId, this.transferJob);
            // A Active notification is already sent in this case
            // But it should be updated in the database
            if (this.status != RFTConstants.STATUS_EXPANDING) {
                this.overallStatus = 
                    this.resourceManager.getOverallStatus();
                if (this.fault != null) {
                    this.overallStatus.setFault(this.fault);
                }

                if (logger.isDebugEnabled()) {
                    logger.debug(this.getTransferIdentifiers()
                            + "setting overall rp to "
                            + "(finished: "
                            + this.overallStatus.getTransfersFinished()
                            + ", active: "
                            + this.overallStatus.getTransfersActive() 
                            + ", failed: "
                            + this.overallStatus.getTransfersFailed()
                            + ", restarted: "
                            + this.overallStatus.getTransfersRestarted()
                            + ", pending: "
                            + this.overallStatus.getTransfersPending()
                            + ", cancelled: "
                            + this.overallStatus.getTransfersCancelled()
                            + ")");
                }

                this.overallStatusRP.set(0, this.overallStatus);
                setRequestStatus();
            }
        } catch (RftDBException e) {
            logger.error(i18n.getMessage("dbUpdateErr") + e.getMessage(), e);
        }
    }

    /**
     * Sets the request status resource property
     */
    private void setRequestStatus() {

        RequestStatusTypeEnumeration reqStatus =
            this.requestStatusType.getRequestStatus();
        int totalTransfers = 0;
        int activeTransfers = this.overallStatus.getTransfersActive();
        int failedTransfers = this.overallStatus.getTransfersFailed();
        int pendingTransfers = this.overallStatus.getTransfersPending();
        int finishedTransfers = this.overallStatus.getTransfersFinished();
        int restartedTransfers = this.overallStatus.getTransfersRestarted();
        totalTransfers = activeTransfers + failedTransfers + pendingTransfers
                + finishedTransfers + restartedTransfers;
        if (activeTransfers > 0 && !reqStatus.getValue().equals("Active")) {
            this.requestStatusType
                    .setRequestStatus(RequestStatusTypeEnumeration.Active);
            this.requestStatusTypeRP.set(0, this.requestStatusType);
        }
        if ((failedTransfers > 0) && !reqStatus.getValue().equals("Failed")) {
            this.requestStatusType
                    .setRequestStatus(RequestStatusTypeEnumeration.Failed);
            this.requestStatusType.setFault(this.fault);
            this.requestStatusTypeRP.set(0, this.requestStatusType);
            return;
        }
        if (totalTransfers == finishedTransfers) {
            if (logger.isDebugEnabled()) {
                logger.debug(this.getTransferIdentifiers()
                             + "setting status RP to Done");
            }
            this.requestStatusType
                    .setRequestStatus(RequestStatusTypeEnumeration.Done);
            this.requestStatusTypeRP.set(0, this.requestStatusType);
            return;
        }
    }

    /*
     * Sets Faults on resource properties
     */
    private void setFault(Exception e) {
        if (e instanceof RftTransientException) {
            this.fault = FaultUtil.createTransferTransientFault(this.requestId,
                    e.getMessage(), e);
            if (logger.isDebugEnabled()) {
                logger.debug("setting transient fault");
            }
            this.requestStatusType.setFault(this.fault);
            this.overallStatus.setFault(this.fault);
        }
        if (e instanceof RftException) {
            this.fault = FaultUtil.createTransferFault(this.requestId,
                    e.getMessage(), e);
            this.requestStatusType.setFault(this.fault);
            this.overallStatus.setFault(this.fault);
        }
        if (e instanceof RftDBException) {
            this.fault = FaultUtil.createRFTDatabaseFault(this.requestId, e);
            this.requestStatusType.setFault(this.fault);
            this.overallStatus.setFault(this.fault);
        }
    }

    /*
     * For Processing states
     */
    private void processStates()
    {
        int transferId = this.transferJob.getTransferId().intValue();

        if (logger.isDebugEnabled())
        {
            logger.debug(this.getTransferIdentifiers()
                         + "processing state for transfer of "
                         + this.transferJob.getSourceUrl() + "  ->  "
                         + this.transferJob.getDestinationUrl());
        }

        switch (this.status)
        {
            case RFTConstants.STATUS_FINISHED:
            {
                logger.debug(
                    this.getTransferIdentifiers() + "transfer done");
                this.statusChanged(RFTConstants.STATUS_FINISHED);
                break;
            }
            case RFTConstants.STATUS_FAILED:
            {
                logger.debug(
                    this.getTransferIdentifiers() + "transfer failed");
                this.statusChanged(RFTConstants.STATUS_FAILED);
                try
                {
                    if (this.dbAdapter.getAllOrNone(this.requestId))
                    {
                        this.dbAdapter.updateToDirty(this.requestId);
                    }
                }
                catch(RftDBException rftdbe)
                {
                    logger.error(i18n.getMessage("dbRetrieveErr"), rftdbe);
                }
                break;
            }
            case RFTConstants.STATUS_RETRYING:
            {
                if (logger.isDebugEnabled())
                {
                    logger.debug(this.getTransferIdentifiers()
                                 + "retry attempt "
                                 + this.transferJob.getAttempts()
                                 + " of "
                                 + this.maxAttempts);
                }

                if (this.transferJob.getAttempts().intValue() >=
                    this.maxAttempts)
                {
                    if (logger.isDebugEnabled())
                    {
                        logger.debug("Transfer " + transferId + ": "
                                     + "transfer failed (retried "
                                     + this.transferJob.getAttempts()
                                     + " times)");
                    }
                    this.statusChanged(RFTConstants.STATUS_FAILED);
                }
                else
                {
                    this.statusChanged(RFTConstants.STATUS_RETRYING);
                }
                break;
            }
            case RFTConstants.STATUS_EXPANDING_DONE:
            {
                if (logger.isDebugEnabled()) {
                    logger.debug(this.getTransferIdentifiers()
                                 + "expanding the url: " 
                                 + this.transferJob.getSourceUrl());
                }
                this.statusChanged(RFTConstants.STATUS_EXPANDING_DONE);
                break;
            }
            default:
            {
                logger.debug("##### Unexpected state #####");
            }
        }
    }

    /**
     * Main processing method for the TransferWork object 
     */
    public void run()
    {
        logger.debug("TransferWork::run() enter");
        Client client = null;
        int transferId = 0, clientType = 0;
        boolean success = false;
        int local_status = this.status;

        try
        {
            transferId = this.transferJob.getTransferId().intValue();
            this.attempts = (this.transferJob.getAttempts().intValue() + 1);
            this.transferJob.setAttempts(new Integer(this.attempts));
            clientType = getClientType(this.transferJob);

            switch(clientType)
            {
                case ClientFactory.DELETE_CLIENT:
                {
                    client = ClientFactory.createDeleteClient(
                        this.transferJob, this.resourceManager,
                        ReliableFileTransferImpl.getConnectionManager());
                    break;
                }
                case ClientFactory.THIRDPARTY_TRANSFER_CLIENT:
                {
                    client = ClientFactory.createThirdPartyTransferClient(
                        this.transferJob, this.resourceManager,
                        ReliableFileTransferImpl.getConnectionManager(),
                        this.rpSet, transferId);
                    break;
                }
                case ClientFactory.URL_EXPANDER_CLIENT:
                {
                    client = ClientFactory.createURLExpanderClient(
                        this.transferJob, this.resourceManager,
                        ReliableFileTransferImpl.getConnectionManager(),
                        this.requestId);
                    break;
                }
                case ClientFactory.GET_PUT_TRANSFER_CLIENT:
                {
                    client = ClientFactory.createGetPutTransferClient(
                        this.transferJob, this.resourceManager,
                        ReliableFileTransferImpl.getConnectionManager(),
                        this.rpSet, transferId);
                    break;
                }
            }

            if (client == null)
            {
                throw new RftException("invalid client type requested");
            }
            statusChanged(RFTConstants.STATUS_ACTIVE);
            client.process();

            if (clientType == ClientFactory.URL_EXPANDER_CLIENT)
            {
                local_status = RFTConstants.STATUS_EXPANDING_DONE;
                this.resourceManager.setNumberOfTransfers(
                    ((URLExpanderClient)client).getNumberOfTransfers());
            }
            else
            {
                local_status = RFTConstants.STATUS_FINISHED;
            }
            success = true;
        }
        catch (RftDBException dbe)
        {
            logger.error(i18n.getMessage("dbRetrieveErr"), dbe);
            local_status =  RFTConstants.STATUS_FAILED;
            setFault(dbe);
        }
        catch (RftException rfe)
        {
            logger.error(i18n.getMessage("terminalTransferErr")
                         + rfe.getMessage(), rfe);
            local_status =  RFTConstants.STATUS_FAILED;
            setFault(rfe);
        }
        catch (RftTransientException rfte)
        {
            logger.error(i18n.getMessage("transientTransferErr")
                         + rfte.getMessage(), rfte);
            local_status = RFTConstants.STATUS_RETRYING;
            setFault(rfte);
        }
        this.status = local_status;
        processStates();
        //startNewTransfer();

        resourceManager.transferCompleted(this.transferJob, success);
        logger.debug("TransferWork::run() exit");
    }

    private String getTransferIdentifiers() {
        return "[Request " + this.requestId
            + ", Transfer " + this.transferJob.getTransferId().intValue()
            + "] ";         
    }

    /**
     * Determine the client type for this request.
     * @param transferJob
     * @return
     */
    private int getClientType(TransferType transferJob)
    {
        int clientType = -1;
        String destUrl = transferJob.getDestinationUrl();
        
        if (destUrl == null)
        {
            clientType = ClientFactory.DELETE_CLIENT;
        } 
        else
        {
            if (transferJob.getStatus().getValue().equals("Dirty"))
            {
                this.transferJob.setSourceUrl(destUrl);
                clientType = ClientFactory.DELETE_CLIENT;
            }
            else
            {
                if (transferJob.getSourceUrl().endsWith("/"))
                {
                    clientType = ClientFactory.URL_EXPANDER_CLIENT;
                }
                else
                {
            	    if (transferJob.getRftOptions().getNotpt().booleanValue())
            	    {
                        clientType = ClientFactory.GET_PUT_TRANSFER_CLIENT;
            	    }
            	    else
            	    {
                        clientType = ClientFactory.THIRDPARTY_TRANSFER_CLIENT;
                    }
                }
            }
        }
        return clientType;
    }
}

