package org.globus.transfer.reliable.service.cache;

import java.util.Date;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.ListIterator;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.transfer.reliable.service.exception.RftException;

/**
 * @Threadsafe
 */
public class ConnectionCacheImpl implements ConnectionCache
{
    private static Log logger =
         LogFactory.getLog(ConnectionCacheImpl.class.getName());

    private ConnectionCounter connectionCounter;
    private LinkedList cache;
    private int type;
    private long entryExpirationTime;
    
    public ConnectionCacheImpl(
           int                type,
           long               entryExpirationTime,
           ConnectionCounter  connectionCounter)
    throws RftException
    {
        this.connectionCounter = connectionCounter;
        this.cache = new LinkedList();
        this.type = type;
        this.entryExpirationTime = entryExpirationTime;
    }

    /**
     * Add a connection to the beginning of the cache.
     */
    public void addConnection(Connection connection)
    {    
        if(logger.isDebugEnabled()) { 
            logger.debug("adding connection to cache");
        }
    	synchronized (cache) {
            cache.addFirst(connection);
        }
        this.connectionCounter.add(connection.getNumberFDs());
    }
    
    /**
     * Get a connection from the cache.
     * @param params The connection must fit to the given params
     * @return a fitting connection or null if no suitable connection was found
     */
    public Connection getConnection(ConnectionParameter params)
    {
        Connection connection = null;
        boolean foundSuitable = false;
        
        // search the cache for a suitable connection
        synchronized (cache) {
            Iterator iter = cache.iterator();
            while (iter.hasNext()) {
                connection = (Connection)iter.next();
                if (connection.checkForReuse(params)) {
                	if(logger.isDebugEnabled()) {
                	    logger.debug("found suitable connection");
                	}
                    iter.remove();
                    foundSuitable = true;
                    break;
                }
            }
        }

        if(foundSuitable) {
            this.connectionCounter.substract(connection.getNumberFDs());
            if (connection.checkConnectionIntegrity()) {
            	return connection;
            } else {
            	if(logger.isDebugEnabled()) {
            	    logger.debug("integrity check failed");
            	}
            	connection.closeConnection();
                /* Idea behind this recursion: the cache is not blocked then.
                   If the expiration time is set to 1 min or so the integrity
                   check should fail very rarely (only when servers are
                   configured to close unused connections quickly
                   [default is 10min]).
                 */
                return getConnection(params);
            }
        }
        return null;
    }

    
    /**
     * Add the last element of the cache to the expired list. The last
     * element is automatically the one that has not been used for the
     * longest time compared to other connections in this cache
     *
     * @return 0 on success, or -1 if the cache cannot be shrunk
     */
    public int shrink()
    {
        Connection tmp;
        synchronized (this.cache)
        {
            if (this.cache.size() > 0)
            {
                tmp = (Connection) cache.removeLast();
            }
            else
            {
                return -1;
            }
        }
        this.connectionCounter.substract(tmp.getNumberFDs());
        tmp.closeConnection();
        return 0;
    }

    /**
     * Get the type of the cache. This is the type of the connection
     * it manages.
     */
    public int getType()
    {
        return this.type;
    }

    /**
     * Get the timestamp of the last (oldest) connection.
     * @return timestamp of the last connection or -1 if cache is empty
     */ 
    public long getOldestTimestamp()
    {
        synchronized(cache) {
            if (cache.size() > 0) {
                return ((Connection)cache.getLast()).getLastUsed();
            }
        }
        return -1;
    }

    /**
     * This method is called when the container goes down and closes
     * all connections, both in the cache and in the connectionToClose
     * list.
     * ConnectionManager registers a ConnectionCacheCleanupHook which
     * calls this method.
     */
    public void closeAllConnections()
    {
    	if(logger.isDebugEnabled()) {
    	    logger.debug("closing all connections");
    	}
        
        synchronized (this.cache) {
            Iterator iter = this.cache.iterator();
            while (iter.hasNext()) {
            	Connection tmp = (Connection)iter.next();
                tmp.closeConnection();
                this.connectionCounter.substract(tmp.getNumberFDs());
                iter.remove();
            }
        }
    }
    
    /**
     * Remove expired connections from the cache.
     * Walk through the cache in reverse order and get all expired connections.
     * The loop is exited if a connection did not expire, because the newest
     * connections are always at the top of the cache.
     */
    public void removeExpiredConnections()
    {
        ListIterator iter;
        Connection tmpConnection;
        LinkedList tmpList = new LinkedList();
        
        // add expired connections to a temporary list
        long now = new Date().getTime();
        synchronized (cache) {
            iter = cache.listIterator(cache.size());
            while (iter.hasPrevious()) {
                tmpConnection = (Connection)iter.previous();
                if(connectionExpired(tmpConnection,now)) {
                    tmpList.add(tmpConnection);
                    iter.remove();
                } else {
                    break;
                }
            }
        }
        
        // close expired connections
        iter = tmpList.listIterator();
        while (iter.hasNext()) {
            Connection tmp = (Connection)iter.next();
            tmp.closeConnection();
            this.connectionCounter.substract(tmp.getNumberFDs());
            iter.remove();
        }
    }

    /**
     * Check if a connection expired
     * @param connection connection to check
     * @return true if the connection expired
     */
    private boolean connectionExpired (
            Connection   connection,
            long         now)
    {
        if ((now - this.entryExpirationTime) > connection.getLastUsed()) {
            return true;
        } else {
            return false;
        }
    }

}
