package org.globus.transfer.reliable.service.client;

import java.net.MalformedURLException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.rft.generated.RFTOptionsType;
import org.globus.rft.generated.TransferType;
import org.globus.transfer.reliable.service.RFTResourceManager;
import org.globus.transfer.reliable.service.cache.ConnectionParameter;
import org.globus.transfer.reliable.service.cache.ConnectionManager;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.globus.util.GlobusURL;
import org.globus.wsrf.ResourcePropertySet;
import org.ietf.jgss.GSSCredential; 

public class ClientFactory
{
    private static Log logger =
        LogFactory.getLog(ClientFactory.class.getName());

    public static final int DELETE_CLIENT = 0;
    public static final int THIRDPARTY_TRANSFER_CLIENT = 1;
    public static final int GET_PUT_TRANSFER_CLIENT = 2;
    public static final int URL_EXPANDER_CLIENT = 3;
    
    /**
     * Create a DeleteClient
     * @param transferJob
     * @param resourceManager
     * @param connectionManager
     * @return
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    public static DeleteClient createDeleteClient(
           TransferType         transferJob,
           RFTResourceManager   resourceManager,
           ConnectionManager    connectionManager)
    throws RftException, RftTransientException, RftDBException 
    {
    	if (transferJob == null) {
    	    throw new RftException("transferJob must not be null here");
    	}
    	if (resourceManager == null) {
    	    throw new RftException("resourceManager must not be null here");
    	}
        if (connectionManager == null) {
            throw new RftException("resourceManager must not be null here");
        }
    	
        if (logger.isDebugEnabled()) {
    	    logger.debug("creating new DeleteClient");
        }
        RFTOptionsType rftOptions = transferJob.getRftOptions();
        GlobusURL sourceUrl = getGlobusURL(transferJob.getSourceUrl());
        GSSCredential credential = resourceManager.getCredential();
        
        DeleteClient client
            = new DeleteClient(
                new ConnectionParameter(
                    sourceUrl.getHost(),
                    sourceUrl.getPort(),
                    sourceUrl.getPath(),
                    getUserDnFromCredential(credential),
                    rftOptions),
                credential,
                connectionManager);
        return client;
    }

    /**
     * Create a GetPutClient for non-3rdParty transfers
     * @param transferJob
     * @param resourceManager
     * @param connectionManager
     * @param rpSet
     * @param transferId
     * @return
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    public static GetPutTransferClient createGetPutTransferClient(
           TransferType         transferJob,
           RFTResourceManager   resourceManager,
           ConnectionManager    connectionManager,
           ResourcePropertySet  rpSet,
           int                  transferId)
    throws RftException, RftTransientException, RftDBException
    {
    	if (transferJob == null) {
    	    throw new RftException("transferJob must not be null here");
    	}
    	if (resourceManager == null) {
    	    throw new RftException("resourceManager must not be null here");
    	}
        if (connectionManager == null) {
            throw new RftException("connectionManager must not be null here");
        }
    	if (rpSet == null) {
    	    throw new RftException(
    	        "ResourcePropertySet must not be null here");
    	}
    	
    	logger.debug("creating new GetPutTransferClient");
    	RFTOptionsType rftOptions = transferJob.getRftOptions();
    	GlobusURL sourceUrl = getGlobusURL(transferJob.getSourceUrl());
        GlobusURL destinationUrl = getGlobusURL(transferJob.getDestinationUrl());
        GSSCredential credential = resourceManager.getCredential();
        String userDN = getUserDnFromCredential(credential);
        
        GetPutTransferClient client
            = new GetPutTransferClient(
                new ConnectionParameter(
                    sourceUrl.getHost(),
                    sourceUrl.getPort(),
                    sourceUrl.getPath(),
                    userDN,
                    rftOptions),
                new ConnectionParameter(
                    destinationUrl.getHost(),
                    destinationUrl.getPort(),
                    destinationUrl.getPath(),
                    userDN,
                    rftOptions),
                credential,
                connectionManager,
                rpSet,
                transferId);
        return client;
    }
    
    /**
     * Create a ThirdPartyTransferClient
     * @param transferJob
     * @param resourceManager
     * @param connectionManager
     * @param rpSet
     * @param transferId
     * @return
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    public static ThirdPartyTransferClient createThirdPartyTransferClient(
           TransferType         transferJob,
           RFTResourceManager   resourceManager,
           ConnectionManager    connectionManager,
           ResourcePropertySet  rpSet,
           int                  transferId)
    throws RftException, RftTransientException, RftDBException
    {
    	if (transferJob == null) {
    	    throw new RftException("transferJob must not be null here");
    	}
    	if (resourceManager == null) {
    	    throw new RftException("resourceManager must not be null here");
    	}
        if (connectionManager == null) {
            throw new RftException("resourceManager must not be null here");
        }
    	if (rpSet == null) {
    	    throw new RftException(
    	        "ResourcePropertySet must not be null here");
    	}

    	logger.debug("creating new ThirdPartyTransferClient");
        RFTOptionsType rftOptions = transferJob.getRftOptions();
        GlobusURL sourceUrl = getGlobusURL(transferJob.getSourceUrl());
        GlobusURL destinationUrl = getGlobusURL(transferJob.getDestinationUrl());
        GSSCredential credential = resourceManager.getCredential();

        ThirdPartyTransferClient client
            = new ThirdPartyTransferClient(
                new ConnectionParameter(
                    sourceUrl.getHost(),
                    sourceUrl.getPort(),
                    sourceUrl.getPath(),
                    destinationUrl.getHost(),
                    destinationUrl.getPort(),
                    destinationUrl.getPath(),
                    getUserDnFromCredential(credential),
                    rftOptions),
                credential,
                connectionManager,
                rpSet,
                transferId);
        return client;
    }

    /**
     * Create a URLExpanderClient
     * @param transferJob
     * @param resourceManager
     * @param connectionManager
     * @param requestId
     * @return
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    public static URLExpanderClient createURLExpanderClient(
           TransferType         transferJob,
           RFTResourceManager   resourceManager,
           ConnectionManager    connectionManager,
           int                  requestId)
    throws RftException, RftTransientException, RftDBException
    {
    	if (transferJob == null) {
    	    throw new RftException("transferJob must not be null here");
    	}
    	if (resourceManager == null) {
    	    throw new RftException("resourceManager must not be null here");
    	}
        if (connectionManager == null) {
            throw new RftException("connectionManager must not be null here");
        }

    	logger.debug("creating new URLExpanderClient");
        RFTOptionsType rftOptions = transferJob.getRftOptions();
        GSSCredential credential = resourceManager.getCredential();        
        GlobusURL sourceUrl
            = getGlobusURL(transferJob.getSourceUrl());
        GlobusURL destinationUrl
            = getGlobusURL(transferJob.getDestinationUrl());

        URLExpanderClient client
            = new URLExpanderClient(
                new ConnectionParameter(
                    sourceUrl.getHost(),
                    sourceUrl.getPort(),
                    sourceUrl.getPath(),
                    getUserDnFromCredential(credential),
                    rftOptions),
                new ConnectionParameter(
                    destinationUrl.getHost(),
                    destinationUrl.getPort(),
                    destinationUrl.getPath(),
                    getUserDnFromCredential(credential),
                    rftOptions),
                resourceManager.getCredential(),
                connectionManager,
                requestId);
        return client;
    }
          
    /**
     * Create an instance of GlobusURL from a String and check
     * if the protocol if gsiftp.
     * @param urlString
     * @return
     * @throws RftException
     */
    public static GlobusURL getGlobusURL(String urlString)
    throws RftException
    {
        GlobusURL url;
        
        if (urlString == null) {
            throw new RftException("Url string is null");
        }
        
        try {
            url = new GlobusURL(urlString);
        } catch (MalformedURLException e) {
            throw new RftException(e);
        }
        
        if (!url.getProtocol().equalsIgnoreCase("gsiftp")) {
            // TODO:  is this explanation ok?
            throw new RftException("Only 3rd party transfers supported");
        }
        return url;
    }
    
    private static String getUserDnFromCredential(GSSCredential credential)
    throws RftException
    {
        String userDN = null;        
        try {
            userDN = credential.getName().toString();
        } catch (Exception e) {
            throw new RftException("Unable to extract the user's DN from "
                + "the credential", e);
        }
        return userDN;
    }
}
